package main

import (
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"github.com/jasonlvhit/gocron"
	"github.com/lamperi/e4bot/spotify"
	"io"
	"log"
	"os"
	"reflect"
	"regexp"
	"strconv"
	"strings"
	"time"
)

type App struct {
	db            *DB
	credentials   Credentials
	spotifyClient *spotify.SpotifyClient
}

type Credentials struct {
	APIURL              string
	UserName            string
	Password            string
	SpotifyClientID     string
	SpotifyClientSecret string
	SpotifyUser         string
	SpotifyCallback     string
	ListenAddr          string
}

func (app *App) CreateSpotifyClient() *spotify.SpotifyClient {
	spotifyClient := spotify.NewClient(app.credentials.SpotifyClientID, app.credentials.SpotifyClientSecret)
	spotifyClient.SetupUserAuthenticate(app.credentials.SpotifyCallback)
	return spotifyClient
}

func (app *App) LaunchWeb() {
	app.spotifyClient = app.CreateSpotifyClient()
	go func() {
		webStart(app.credentials.ListenAddr, app.db, app.spotifyClient)
	}()
}

func appendSong(wikiText string, author string, newSong string) string {
	const AUTHOR_MARK = "<!-- Lisääjä -->"
	const SONG_MARK = "<!-- Kappale -->"
	lines := strings.Split(wikiText, "\n")
	authorPrevIndex := -2

	changedLines := make([]string, 0, len(lines))
	for index, line := range lines {
		if strings.Index(line, AUTHOR_MARK) != -1 && strings.Index(line, author) != -1 {
			authorPrevIndex = index
		}
		if authorPrevIndex == (index-1) && strings.Index(line, SONG_MARK) != -1 {
			changedLines = append(changedLines, "| "+SONG_MARK+" "+newSong)
		} else {
			changedLines = append(changedLines, line)
		}
	}
	return strings.Join(changedLines, "\n")
}

func (app *App) wikiClient() *WikiClient {
	wiki := CreateWikiClient(app.credentials.APIURL, app.credentials.UserName, app.credentials.Password)
	return wiki
}

func (app *App) AddSong(updateTitle, updateSection, author, song string) (bool, error) {
	wiki := app.wikiClient()

	sections, err := wiki.GetWikiPageSections(updateTitle)
	if err != nil {
		return false, err
	}

	for _, section := range sections {
		if updateSection == section.title {
			wikiText, err := wiki.GetWikiPageSectionText(updateTitle, section.index)
			if err != nil {
				return false, err
			}
			changedWikiText := appendSong(wikiText, author, song)

			if false {
				// Stub
				fmt.Println("Pretend to update wiki text to ", updateTitle, section.index, changedWikiText,
					fmt.Sprintf("Added %s song for %s", updateSection, author))
				return true, nil
			}

			return wiki.EditWikiPageSection(updateTitle, section.index, changedWikiText,
				fmt.Sprintf("Added %s song for %s", updateSection, author))
		}
	}
	return false, errors.New("Could not find matching section")
}

func (app *App) SongSynced(userId, roundId int) error {
	_, err := app.db.EntrySynced(userId, roundId)
	return err
}

func (app *App) SubmitSongs() {
	entries, err := app.db.FindEntriesToSync()
	if err != nil {
		log.Println("Error while finding entries to sync:", err)
		return
	}

	for _, entry := range entries {
		song := songWikiText(entry.spotifyURL, entry.artist, entry.title)

		fmt.Println("Time has passed for " + song)

		success, err := app.AddSong(entry.article, entry.section, entry.username, song)
		if err != nil {
			log.Println("Error while adding song:", err)
		}
		if success {
			err = app.SongSynced(entry.userId, entry.roundId)
			if err != nil {
				fmt.Println("Error received:", err)
			}
		}
	}
}

func isCurrentAuthor(line, author string) bool {
	authorIndex := strings.Index(line, author)
	endIndex := strings.Index(line, "-->")
	return authorIndex != -1 && authorIndex < endIndex
}

func parseScore(line string) string {
	parts := strings.Split(line, "-->")
	if len(parts) < 2 {
		return ""
	}
	score := strings.TrimRight(strings.Trim(parts[1], " \t\n"), "p")
	if score == "" {
		return score
	}
	rating, _ := regexp.Compile("^\\{\\{Rating\\|\\d([.,]\\d)\\|5\\}\\}?$")
	if rating.MatchString(score) {
		score = strings.Replace(score, "{{Rating|", "", 1)
		score = strings.Replace(score, "|5}}", "", 1)
	}
	number, _ := regexp.Compile("^\\d([.,]\\d)?$")
	if number.MatchString(score) {
		return strings.Replace(score, ",", ".", 1)
	}
	numberHalf, _ := regexp.Compile("^\\d½$")
	if numberHalf.MatchString(score) {
		return fmt.Sprintf("%c.5", score[0])
	}
	stars, _ := regexp.Compile("^\\*+$")
	if stars.MatchString(score) {
		return fmt.Sprintf("%d", len(score))
	}
	imageStars, _ := regexp.Compile("^(\\[\\[Image:[01]\\.png\\]\\]){5}$")
	if imageStars.MatchString(score) {
		return fmt.Sprintf("%d", strings.Count(score, "1"))
	}
	quarterScore, _ := regexp.Compile("^\\d-\\d½?$")
	if quarterScore.MatchString(score) {
		return fmt.Sprintf("%c.25", score[0])
	}
	thirdQuarterScore, _ := regexp.Compile("^\\d½?-\\d$")
	if thirdQuarterScore.MatchString(score) {
		return fmt.Sprintf("%c.75", score[0])
	}
	fmt.Printf("Could not match '%s'\n", score)
	return ""
}

func appendAverages(wikiText string) string {
	const AUTHOR_MARK = "<!-- Lisääjä -->"
	const SONG_MARK = "<!-- Kappale -->"
	const AVERAGE_MARK = "<!-- KA -->"
	lines := strings.Split(wikiText, "\n")
	isScore := false
	scores := make([]string, 0)
	count := 0
	currentAuthor := ""

	changedLines := make([]string, 0, len(lines))
	for _, line := range lines {
		if strings.Index(line, AUTHOR_MARK) != -1 {
			currentAuthor = strings.Trim(strings.Split(line, AUTHOR_MARK)[1], " \t")
		} else if strings.Index(line, SONG_MARK) != -1 {
			isScore = true
			scores = make([]string, 0)
			count = 0
		} else if isScore && strings.Index(line, AVERAGE_MARK) == -1 {
			if !isCurrentAuthor(line, currentAuthor) {
				score := parseScore(line)
				if score != "" {
					scores = append(scores, score)
					count += 1
				} else {
					scores = append(scores, "0")
				}

			}
		}
		if strings.Index(line, AVERAGE_MARK) != -1 && count > 2 {
			expression := fmt.Sprintf("'''{{#expr:(%s)/%d round 2}}'''", strings.Join(scores, "+"), count)
			newLine := "| " + AVERAGE_MARK + " " + expression
			changedLines = append(changedLines, newLine)
			if newLine != line {
				fmt.Printf("Difference for %s\n%s\n%s\n", currentAuthor, newLine, line)
			}

		} else {
			changedLines = append(changedLines, line)
		}
	}
	return strings.Join(changedLines, "\n")
}

func findPlaylist(wikiText string) (string, []string) {
	const SONG_MARK = "<!-- Kappale -->"
	const SPOTIFY_MARK = "https://open.spotify.com/track/"
	const SPOTIFY_PLAYLIST_MARK = "/playlist/"
	const PLAYLIST_MARK = " Spotify-soittolista]"
	lines := strings.Split(wikiText, "\n")

	playlistId := ""
	tracks := make([]string, 0)
	for _, line := range lines {
		if strings.Index(line, SONG_MARK) != -1 {
			i := strings.Index(line, SPOTIFY_MARK)
			if i != -1 {
				j := strings.Index(line[i:], " ")
				if j != -1 {
					j += i
				}
				trackId := line[i+len(SPOTIFY_MARK) : j]
				tracks = append(tracks, trackId)
			}
		} else if strings.Index(line, SPOTIFY_PLAYLIST_MARK) != -1 && strings.Index(line, PLAYLIST_MARK) != -1 {
			i := strings.Index(line, SPOTIFY_PLAYLIST_MARK)
			j := strings.Index(line[i:], PLAYLIST_MARK)

			playlistId = line[i+len(SPOTIFY_PLAYLIST_MARK) : i+j]
			q := strings.Index(playlistId, "?")
			if q != -1 {
				playlistId = playlistId[:q]
			}
		}
	}
	fmt.Printf("Found playlist %s and tracks %s\n", playlistId, tracks)

	return playlistId, tracks
}

func appendPlaylist(wikiText string, playlist *spotify.PlaylistInfo) string {
	changedText := wikiText + `
	[` + playlist.ExternalUrls.Spotify + ` Spotify-soittolista]
	`
	return changedText
}

func (app *App) AutomateSection(title string) error {
	wiki := app.wikiClient()

	sections, err := wiki.GetWikiPageSections(title)
	if err != nil {
		return err
	}

	_, currentWeek := time.Now().ISOWeek()
	numberReg, _ := regexp.Compile("\\d+")
	for _, section := range sections {
		weekStr := numberReg.FindString(section.title)
		if weekStr != "" {
			weekNumber, _ := strconv.Atoi(weekStr)
			if weekNumber < currentWeek-1 {
				continue
			}
			if weekNumber > currentWeek {
				break
			}
			fmt.Println("Checking section", section.title)

			wikiText, err := wiki.GetWikiPageSectionText(title, section.index)
			if err != nil {
				return err
			}
			message := ""
			changedWikiText := appendAverages(wikiText)
			if changedWikiText != wikiText {
				message = message + fmt.Sprintf("Calculate averages for week %d. ", weekNumber)
			}

			if app.spotifyClient.HasUserLogin() {
				playlistId, tracks := findPlaylist(changedWikiText)
				currentTracks, err := app.db.FindPlaylistBySection(section.title)
				if len(tracks) > 0 && (err != nil || reflect.DeepEqual(currentTracks, tracks)) {

					spotify := app.spotifyClient
					if playlistId == "" {

						info, err := spotify.NewPlaylist(title+" "+section.title, app.credentials.SpotifyUser)
						if err != nil {
							log.Println("Error creating playlist")
							return err
						}
						playlistId = info.Id
						changedWikiText = appendPlaylist(changedWikiText, info)
						message = message + fmt.Sprintf("Added link to Spotify playlist for week %d.", weekNumber)
					}
					err := spotify.UpdatePlaylist(app.credentials.SpotifyUser, playlistId, tracks)
					if err != nil {
						log.Println("Error updating playlist")
						return err
					}
					_, err = app.db.UpdatePlaylistBySection(section.title, tracks)
					if err != nil {
						return err
					}
				}

			}

			if message != "" {

				_, err := wiki.EditWikiPageSection(title, section.index, changedWikiText,
					fmt.Sprintf("Calculate averages for week %d", weekNumber))
				err = nil
				if err != nil {
					return err
				}
			}
		}
	}
	return nil
}

func (app *App) AutomateSectionTask() {
	panels, err := app.db.FindAllPanels()
	if err != nil {
		fmt.Println("Error while checking db for panels:", err)
		return
	}
	for _, panel := range panels {
		fmt.Println("Checking panel", panel)
		err := app.AutomateSection(panel)
		if err != nil {
			fmt.Println("Error while processing panel:", err)
		}
	}

}

func initCreds() (Credentials, error) {
	var credsFile string
	flag.StringVar(&credsFile, "credentials", "credentials.json", "JSON config to hold app credentials")
	flag.Parse()

	var credentials Credentials
	f, err := os.Open(credsFile)
	if err != nil {
		return credentials, err
	}
	defer f.Close()

	if err != nil {
		log.Fatal(err)
		return credentials, err
	}
	dec := json.NewDecoder(f)
	for {
		if err := dec.Decode(&credentials); err == io.EOF {
			break
		} else if err != nil {
			log.Fatal(err)
		}
	}
	return credentials, nil
}

func songWikiText(url string, artist string, title string) string {
	return "[" + url + " " + artist + " - " + title + "]"
}

func main() {
	creds, err := initCreds()
	if err != nil {
		panic(err)
	}

	a := App{InitDatabase(), creds, nil}
	a.LaunchWeb()

	gocron.Every(1).Hour().Do(a.AutomateSectionTask)
	gocron.Every(1).Second().Do(a.SubmitSongs)
	<-gocron.Start()

}
